from ATOM3Float import ATOM3Float
import threading
from time import sleep

from CBD_Constants import *
from CBD_DepGraph import DepGraph
from CBD_PlotterController import CBD_PlotterController

class TS_Simulator(threading.Thread):
  """ Simulator for Causal Block Diagram models
  """

  def __init__(self, model, curIteration=0):
    self.__running      = False
    self.__model        = model
    self.__curIteration = curIteration    
    threading.Thread.__init__(self, name='simulator')
    
  def __getBlocks(self):
    return self.__model.listNodes[BLOCK]
    
  def __getSignals(self, signalTypes=[]):
    if signalTypes == []: #empty list, return all signals
      signalTypes = [BLOCK_SIGNAL, TIME_SIGNAL, TERMINATE_SIGNAL]
    answer = []
    for signalType in signalTypes:
      if self.__model.listNodes.has_key(signalType):
        answer += self.__model.listNodes[signalType]
      else:
        raise ValueError("Incorrect Signal Type")
    return answer        
    
  def __getSimControlBlock(self): 
    simControl=self.__model.listNodes[SIMCONTROL]
    assert len(simControl)==1, "There should be only one simulator control block"
    return simControl[0]
  
  def __getInfluencerBlock(self, signal):
    assert len(signal.in_connections_)==1, "Model syntax error: signal "+\
                                           str(signal.signal_name)+\
                                           " must have exactly"+\
                                           " one influencer block"
    return signal.in_connections_[0]

  def __getDependentBlock(self, signal):
    assert len(signal.out_connections_)==1, "Model syntax error: signal "+\
                                           str(signal.signal_name)+\
                                           " must have exactly"+\
                                           " one dependent block"
    return signal.out_connections_[0] 
  
  def getBlockType(self, block):
    if self.__getType(block)==BLOCK:
      index = block.block_type.getValue()[1]
      return block.block_type.getValue()[0][index]
    else:
      raise ValueError("Expected a Block node")
    
  def getBlockName(self, block):
    if self.__getType(block)==BLOCK:
      return block.block_name.getValue()
    else:
      raise ValueError("Expected a Block node")

  def __getType(self, ASGNode):
    return ASGNode.getTypeName()

  # Get the signals of a block that are neither IC, FALSE, or TRUE input ports
  def getInputSignals(self, block):
    answer = []
    for signal in block.in_connections_:
      if (signal not in block.block_IC_port and signal not in block.block_TRUE_in_port and signal not in block.block_FALSE_in_port):
        answer.append(signal)
    return answer 

  def __getInputValues(self, block):  
    if self.__getType(block)==BLOCK:
      inputs=[]
      for signal in block.in_connections_:        
        inputs.append(signal.signal.getValue()[-1])        
    else:
      raise ValueError("Expected a Block node")
    
  def __getCurrentDiscreteTime(self):
    # Semantically, this may not make sense, but we use the curIteration as out discrete time base,
    # so that a non-0 initial value for simControl.simControl_discreteTime can be used.
    return self.__curIteration
    
  def __getCurrentContinuousTime(self):
    return self.__getSimControlBlock().simControl_continuousTime.getValue()
    
  def __getTimeIncrement(self):
    return self.__getSimControlBlock().simControl_timeIncrement.getValue()

  def __getTermSignal(self):
    simControl = self.__getSimControlBlock()
    assert len(simControl.simControl_terminateMonitor_port) == 1, "There must only be one input to the terminate signal"
    return simControl.simControl_terminateMonitor_port[0].signal.getValue()[self.__getCurrentDiscreteTime()].getValue()

  # Traverses the CBD and checks that all blocks are legally connected
  def checkLegality(self):
    blocks = self.__getBlocks()
    for block in blocks:
      blockType = self.getBlockType(block)
      blockName = self.getBlockName(block)
      if (blockName == ""):
         blockName = blockType
      # for negator, inverter, delay, derivative, integrator, and test blocks, one and only one input signal is allowed
      if (blockType == NEGATOR_BLOCK or blockType == INVERTER_BLOCK or blockType == DELAY_BLOCK or blockType == DERIVATIVE_BLOCK or blockType == INTEGRATOR_BLOCK or blockType == TEST_BLOCK):
        assert len(self.getInputSignals(block)) == 1, blockName + " has invalid number of inputs"
      # for delay, derivative and integrator blocks, one and only one IC input is allowed
      if (blockType == DELAY_BLOCK or blockType == DERIVATIVE_BLOCK or blockType == INTEGRATOR_BLOCK):
        assert len(block.block_IC_port) == 1, blockName + " has invalid number of initial conditions"
      # for test blocks, one and only one TRUE input is allowed, and one and only one FALSE input is allowed
      if (blockType == TEST_BLOCK):
        assert len(block.block_TRUE_in_port) == 1, blockName + " has invalid number of inputs (TRUE input)"
        assert len(block.block_FALSE_in_port) == 1, blockName + " has invalid number of inputs (FALSE input)"

  # Constructs the dependency graph of the current state of the CBD
  def __createDepGraph(self):
    blocks = self.__getBlocks()
    depGraph = DepGraph()

    for block in blocks:      
      depGraph.addMember(block)

    for block in blocks:
      blockType = self.getBlockType(block)
      # Treat dependencies on time-delay blocks differently: if the current time is 0, then the block only depends on
      # the IC; else, it depends only on the input signal.
      if ((blockType == DERIVATIVE_BLOCK) or (blockType == INTEGRATOR_BLOCK) or (blockType == DELAY_BLOCK)):
        if (self.__getCurrentDiscreteTime() == 0):
          # Before setting a dependency, we must check if the dependent is really a block, and not the simControl
          if (block.block_IC_port[0].in_connections_[0] in blocks):
            depGraph.setDependency(block, block.block_IC_port[0].in_connections_[0])
        else:
          if (blockType == DERIVATIVE_BLOCK):
            inputb = self.getInputSignals(block)[0].in_connections_[0]
            if (inputb in blocks):
              depGraph.setDependency(block, inputb)
      else:
        for link in block.in_connections_:
          if (link.in_connections_[0] in blocks):
            depGraph.setDependency(block, link.in_connections_[0])
    return depGraph

  # Takes care of pushing a block's output value on all its outwards connections
  def __pushOutputValue(self, block, value):
    if self.__getType(block)==BLOCK:
      for signal in block.out_connections_:        
        signal.signal.getValue().append(value)       
    else:
      raise ValueError("Expected a Block node") 

  # Pushes time on the connected signals
  def __pushTime(self):
    simControl     = self.__getSimControlBlock()
    discreteTime   = simControl.simControl_discreteTime.getValue()
    continuousTime = simControl.simControl_continuousTime.getValue()
    timeIncr       = simControl.simControl_timeIncrement.getValue()      
    
    for signal in simControl.simControl_discreteTime_port:
      signal.signal.getValue().append(ATOM3Float(discreteTime))
     
    for signal in simControl.simControl_continuousTime_port:
      signal.signal.getValue().append(ATOM3Float(continuousTime))
      
    for signal in simControl.simControl_timeIncrement_port:
      signal.signal.getValue().append(ATOM3Float(timeIncr))

  # Increment discrete time by 1, and continuous time by the time increment
  def __incrTime(self):
    simControl     = self.__getSimControlBlock()
    discreteTime   = simControl.simControl_discreteTime.getValue()
    continuousTime = simControl.simControl_continuousTime.getValue()
    timeIncr       = simControl.simControl_timeIncrement.getValue() 
    simControl.simControl_discreteTime.setValue(discreteTime+1)
    simControl.simControl_continuousTime.setValue((discreteTime+1)*timeIncr)
    self.__curIteration += 1

  # Dump some info about the depGraph for debugging purposes
  def __dumpDependencies(self, depGraph):
    print "discrete_time = ", self.__getCurrentDiscreteTime()
    print "current_interation = ", self.__curIteration
    blocks = self.__getBlocks()
    for block in blocks:
      print "Block: ", self.getBlockName(block)
      print "Dependents: "
      for dependent in depGraph.getDependents(block):
        print "  ", self.getBlockName(dependent)
      print "Influencers: "
      for influencer in depGraph.getInfluencers(block):
        print "  ", self.getBlockName(influencer)

  # Computes the output values for all blocks in the CBD, in topSort order
  # Delegates to the appropriate function according to the block type
  def __computeBlocks(self, depGraph, sortedGraph):
    for component in sortedGraph:
      if (not self.__hasCycle(depGraph, component)):
        block = component[0]    
        blockType = self.getBlockType(block)
        if (blockType == GENERIC_BLOCK):
          self.__computeGeneric(block)
        elif (blockType == CONSTANT_BLOCK):
          self.__computeConstant(block)
        elif (blockType == NEGATOR_BLOCK):
          self.__computeNegator(block)
        elif (blockType == INVERTER_BLOCK):
          self.__computeInverter(block)
        elif (blockType == ADDER_BLOCK):
          self.__computeAdder(block)
        elif (blockType == PRODUCT_BLOCK):
          self.__computeProduct(block)
        elif (blockType == DELAY_BLOCK):
          self.__computeDelay(block)
        elif (blockType == DERIVATIVE_BLOCK):
          self.__computeDerivative(block)
        elif (blockType == INTEGRATOR_BLOCK):
          self.__computeIntegrator(block)
        elif (blockType == TEST_BLOCK):
          self.__computeTest(block)
        elif (blockType == FILEIO_BLOCK):
          self.__computeFileIO(block)
      else:
        # Detected a strong component
        assert self.__isLinear(component), "Cannot solve non-linear algebraic loop"
        solverInput = self.__constructLinearInput(component)
        self.__pushDummyValuesOutOfLoop(component) # the linear solver would usually go here
        self.__printLinearInputMatrices(component, solverInput)
  
  def __computeGeneric(self, block):
    raise ValueError("Generic block found! Please specify a type.")

  def __computeConstant(self, block):
    self.__pushOutputValue(block, ATOM3Float(block.block_out_value.getValue()))

  def __computeNegator(self, block):
    self.__pushOutputValue(block, ATOM3Float(-block.in_connections_[0].signal.getValue()[self.__getCurrentDiscreteTime()].getValue()))

  def __computeInverter(self, block):
    self.__pushOutputValue(block, ATOM3Float(1/block.in_connections_[0].signal.getValue()[self.__getCurrentDiscreteTime()].getValue()))
    
  def __computeAdder(self, block):
    sum = 0
    for in_link in block.in_connections_:
      sum += in_link.signal.getValue()[self.__getCurrentDiscreteTime()].getValue()
    self.__pushOutputValue(block, ATOM3Float(sum))

  def __computeProduct(self, block):
    product = 1
    for in_link in block.in_connections_:
      product *= in_link.signal.getValue()[self.__getCurrentDiscreteTime()].getValue()
    self.__pushOutputValue(block, ATOM3Float(product))

  def __computeDelay(self, block):
    output = 0
    if (self.__getCurrentDiscreteTime() == 0):
      output = block.block_IC_port[0].signal.getValue()[self.__getCurrentDiscreteTime()].getValue()
    else:
      output = self.getInputSignals(block)[0].signal.getValue()[self.__getCurrentDiscreteTime()-1].getValue()
    self.__pushOutputValue(block, ATOM3Float(output))

  def __computeDerivative(self, block):
    output = 0
    if (self.__getCurrentDiscreteTime() == 0):
      output = block.block_IC_port[0].signal.getValue()[self.__getCurrentDiscreteTime()].getValue()
    else:
      output = (self.getInputSignals(block)[0].signal.getValue()[self.__getCurrentDiscreteTime()].getValue() - self.getInputSignals(block)[0].signal.getValue()[self.__getCurrentDiscreteTime()-1].getValue()) / self.__getTimeIncrement()
    self.__pushOutputValue(block, ATOM3Float(output))

  def __computeIntegrator(self, block):
    output = 0
    if (self.__getCurrentDiscreteTime() == 0):
      output = block.block_IC_port[0].signal.getValue()[self.__getCurrentDiscreteTime()].getValue()
    else:
      output = block.out_connections_[0].signal.getValue()[self.__getCurrentDiscreteTime()-1].getValue() + self.__getTimeIncrement() * self.getInputSignals(block)[0].signal.getValue()[self.__getCurrentDiscreteTime()-1].getValue()
    self.__pushOutputValue(block, ATOM3Float(output))

  def __computeTest(self, block):
    output = 0
    if (self.getInputSignals(block)[0].signal.getValue()[self.__getCurrentDiscreteTime()].getValue() > 0):
      output = block.block_TRUE_in_port[0].signal.getValue()[self.__getCurrentDiscreteTime()].getValue()
    else:
      output = block.block_FALSE_in_port[0].signal.getValue()[self.__getCurrentDiscreteTime()].getValue()
    self.__pushOutputValue(block, ATOM3Float(output))

  def __computeFileIO(self, block):
    raise ValueError("Cannot treat FileIO blocks")

  # Function used to detect cycles
  def __hasCycle(self, depGraph, component):
    assert len(component)>=1, "A component should have at least one element"
    if len(component)>1:
      return True
    else:     
      if depGraph.hasDependency(component[0], component[0]):
        return True
      else:
        return False

  # Determines if an algebraic loop describes a linear equation or not
  def __isLinear(self, strongComponent):
    for block in strongComponent:
      blockType = self.getBlockType(block)
      if (not (blockType == ADDER_BLOCK or blockType == NEGATOR_BLOCK)):
        if (blockType == PRODUCT_BLOCK):
          alreadyHasOneInputInLoop = False
          for inputSignal in block.in_connections_:
            if (inputSignal.in_connections_[0] in strongComponent):
              if (alreadyHasOneInputInLoop):
                      print "Error:", self.getBlockType(block), self.getBlockName(block), "has more than one input from a loop"
                      return False
              else:
                alreadyHasOneInputInLoop = True
        else:
          print "Error:", self.getBlockType(block), self.getBlockName(block), "was detected in a loop"
          return False
    print "Linear algebraic loop detected"
    return True

  # Constructs input for a solver of systems of linear equations
  # Input consists of two matrices:
  # M1: coefficient matrix, where each row represents an equation of the system
  # M2: result matrix, where each element is the result for the corresponding equation in M1
  def __constructLinearInput(self, strongComponent):
    size = len(strongComponent)
    row = []
    M1 = [] 
    M2 = []
    
    # Initialize matrices with zeros
    i = 0
    while (i < size):
      j = 0
      row = []
      while (j < size):
        row.append(0)
        j += 1
      M1.append(row)
      M2.append(0)
      i += 1
    
    for block in strongComponent:
      blockIndex = strongComponent.index(block) 
      M1[blockIndex][blockIndex] = 1
      blockType = self.getBlockType(block)
      if (blockType == ADDER_BLOCK):
        for link in block.in_connections_:
          if (link.in_connections_[0] in strongComponent):
            M1[blockIndex][strongComponent.index(link.in_connections_[0])] = -1
          else:
            M2[blockIndex] += link.signal.getValue()[self.__getCurrentDiscreteTime()].getValue()
      elif (blockType == NEGATOR_BLOCK):
        M1[blockIndex][strongComponent.index(block.in_connections_[0].in_connections_[0])] = 1
      elif (blockType == PRODUCT_BLOCK):
        product = -1
        in_link = block.in_connections_[0].in_connections_[0]
        for link in block.in_connections_:
          if (link.in_connections_[0] not in strongComponent):
            product = product * link.signal.getValue()[self.__getCurrentDiscreteTime()].getValue()
          else:
            in_link = link
        M1[blockIndex][strongComponent.index(in_link.in_connections_[0])] = product
    return [M1, M2]

  # Function to print out the input matrices in a formatted way
  def __printLinearInputMatrices(self, strongComponent, solverInput):
    print "Indeces:"
    i = 0
    for block in strongComponent:
      print i, ":", self.getBlockType(block), self.getBlockName(block)
      i += 1

    print "Matrices:"
    M1 = solverInput[0]
    M2 = solverInput[1]
    for row in M1:
      print row, "=", M2[M1.index(row)]

  # Instrumental function that pushes a dummy value out of the outputs of a loop
  # Do so because we don't have a linear solver yet; but, for the simulation to
  # continue, we need values coming out of the loop!!!
  def __pushDummyValuesOutOfLoop(self, strongComponent):
    dummy = ATOM3Float(999)
    for block in strongComponent:
      self.__pushOutputValue(block, dummy)

  # Main loop function                         
  def __simulate(self):
    maxIteration = self.__getSimControlBlock().simControl_maxIterations.getValue()
    
    self.checkLegality()

    print "Iteration ", self.__getCurrentDiscreteTime()
    self.__pushTime()

    depGraph = self.__createDepGraph()
    sortedGraph = depGraph.getStrongComponents()
    self.__computeBlocks(depGraph, sortedGraph)

    self.__incrTime()

    depGraph = self.__createDepGraph()
    sortedGraph = depGraph.getStrongComponents()

    terminate = 0
    while self.__running and self.__curIteration < maxIteration and not terminate:       
      sleep(0.2)
      print "Iteration ", self.__getCurrentDiscreteTime()
      self.__pushTime()
      self.__computeBlocks(depGraph, sortedGraph)
      terminate = (self.__getTermSignal() > 0)
      self.__incrTime()
      CBD_PlotterController.getInstance().notify()
    self.stopSimulation()
            
  def stopSimulation(self):
    self.__running = False        
           
  def run(self):
    self.__running=True
    self.__simulate()
    
  def getCurIter(self):
    return self.__curIteration     
                           
#wrapper for class methods.            
class Callable:
    def __init__(self, anycallable):
        self.__call__ = anycallable      
      
class SimulatorManager: 
  """Singleton class managing the simulator
  """
  instance = None

  def __init__(self, model):
    if SimulatorManager.instance!=None:
      raise RunTimeError("Singleton SimulatorManager is already instantiated")
    #simulator wrapper class only instantiated once:        
    self.__model        = model
    self.__simulator    = None
    self.__curIteration = 0
             
  def __clearSignals(self):
    signalTypes = [BLOCK_SIGNAL, TIME_SIGNAL, TERMINATE_SIGNAL]
    signals = []
    for signalType in signalTypes:
      if self.__model.listNodes.has_key(signalType):
        signals+= self.__model.listNodes[signalType]
      else:
        raise ValueError("Incorrect Signal Type '"+str(signalType)+"'")   
    for signal in signals:
      signal.signal.setValue([]) 
                  
  def __resetTime(self):
    simControl=self.__model.listNodes[SIMCONTROL]
    assert len(simControl)==1, "There should be only one simulator control block"
    simControl=simControl[0]  
    
    simControl.simControl_discreteTime.setValue(0)
    simControl.simControl_continuousTime.setValue(0)      
    
  def getInstance(model):
    if SimulatorManager.instance == None:    
      SimulatorManager.instance = SimulatorManager(model)
    return SimulatorManager.instance             
  getInstance=Callable(getInstance)
  
  def isRunning(self):
    return self.__simulator != None
  
  def startResume(self): 
    """ Called when the Play button is pressed
    """
    if not self.isRunning(): 
      if self.__curIteration==0:
        print "Starting the simulation"
      else:
        print "Resuming the simulation"
      self.__simulator = TS_Simulator(self.__model, self.__curIteration)      
      self.__simulator.start()
     
  def pauseSimulation(self): 
    """ Called when the Pause button is pressed
    """
    if self.isRunning():
      print "Pausing the simulation"
      self.__curIteration = self.__simulator.getCurIter()
      self.__simulator.stopSimulation()
      self.__simulator.join()
      self.__simulator = None     
    
  def stopSimulation(self):     
    """ Called when the Rewind button is pressed
    """
    if self.isRunning():
      print "Stopping the simulation"
      self.__simulator.stopSimulation()  
      self.__simulator.join()
      self.__simulator = None 
    self.__curIteration = 0
    self.__clearSignals()
    self.__resetTime()
                        
