# Classes for blocks


class BaseBlock:
  """
  BaseBlock is the base class of all other specific types of blocks,
  and also represents the following types of blocks:
    Adder | Negator | Product | Inverter
  Unlike other blocks, they only have common attributes:
    block name, block type, signals, links
  """

  def __init__(self, block_name, block_type):
    """
    Constructor
    """
    self.__block_name = block_name
    self.__block_type = block_type
    
    # list of signals generated by this block
    self.signals = []     
    
    # list of all other blocks linking into this block;
    # each element of the list is a tuple of four elements: 
    # the link-in block object, the name of the connection between these two blocks, 
    # the name of "inport_name" and the name of "outport_name" 
    # (the last two elements are meaningful only for the connection with a ChildBlock).
    # e.g., block A has only one IN connection which is from block B with name "s1",
    # then in order to refer to B, we do like this:
    # (blockB, s, inputn, outputn) = blockA.linksIN[0]
    self.linksIN = []
  
  def getBlockName(self):
    return self.__block_name
    
  def getBlockType(self):
    return self.__block_type
  
  def dumpSignals(self):
    """
    print out all signal values generated by the block during simulation
    """
    print self.__block_name + " " + self.__block_type + \
          " " + str(self.signals)
  
  def dumpBlock(self):
    """
    print out the basic information of the block;
    for specific types of blocks, they override this method to give specific information about the block
    """
    rtstr = "Block :: " + self.__block_name + \
            "@" + self.__block_type
    for link in self.linksIN:
      rtstr += "\n" + "  IN Block :: " + link[0].getBlockName() + \
               "@" + link[0].getBlockType() 
    print rtstr
    
  def __str__(self):
    rtstr = "Block :: " + self.__block_name + "@" + self.__block_type  + "\n"
    for link in self.linksIN:
      rtstr += "  IN Block :: " + link[0].getBlockName() + "@" + link[0].getBlockType() + "\n"
    return rtstr
  
  def __repr__(self):
    rtstr = "Block :: " + self.__block_name + "@" + self.__block_type  + "\n"
    for link in self.linksIN:
      rtstr += "  IN Block :: " + link[0].getBlockName() + "@" + link[0].getBlockType() + "\n"
    return rtstr
    
  def resetBlockName(self, prefix):
    self.__block_name = prefix + "." + self.__block_name
    
    
class ConstantBlock(BaseBlock):
  """
  ConstantBlock represents Constant blocks which has a special attribute "init_value"
  to store the constant value.
  """
  
  def __init__(self, block_name, block_type, init_value):
    BaseBlock.__init__(self, block_name, block_type)
    self.init_value = init_value
    
  def __str__(self):    
    return BaseBlock.__str__(self) + "  Initial Value :: " + str(self.init_value) + "\n"
  
  def __repr__(self):    
    return BaseBlock.__repr__(self) + "  Initial Value :: " + str(self.init_value) + "\n"
    
  def dumpBlock(self):
    BaseBlock.dumpBlock(self)
    print "  Initial Value :: " + str(self.init_value)
    
    
class GenericBlock(BaseBlock):
  """
  GenericBlock represents Generic blocks which has a special attribute "block_operator"
  to store the name of a math operator.
  So far, we only support methods in math module of Python library which has only one parameter
  (http://docs.python.org/lib/module-math.html)
  e.g., sin, cos.
  """
  
  def __init__(self, block_name, block_type, block_operator):
    BaseBlock.__init__(self, block_name, block_type)
    self.block_operator = block_operator
    
  def __str__(self):    
    return BaseBlock.__str__(self) + "  Operator :: " + self.block_operator + "\n"
  
  def __repr__(self):    
    return BaseBlock.__repr__(self) + "  Operator :: " + self.block_operator + "\n"
    
  def dumpBlock(self):
    BaseBlock.dumpBlock(self)
    print "  Operator :: " + self.block_operator
    
    
class TimeDelayBlock(BaseBlock):
  """
  TimeDelayBlock represents Delay, Integrator and Derivative blocks which has two special attributes:
  "IC" and "timeIncrement". 
  """
  
  def __init__(self, block_name, block_type, time_increment):
    BaseBlock.__init__(self, block_name, block_type)
    # a tuple of the link-in IC block object and the name of the connection between them
    self.IC = None
    
    # time increment for time delay blocks
    self.timeIncrement = time_increment
    
  def __str__(self):    
    return BaseBlock.__str__(self) + \
           "  IC Block :: " + self.IC[0].getBlockName() + "@" + self.IC[0].getBlockType() + "\n"
  
  def __repr__(self):
    return BaseBlock.__repr__(self) + \
           "  IC Block :: " + self.IC[0].getBlockName() + "@" + self.IC[0].getBlockType() + "\n"
    
  def dumpBlock(self):
    BaseBlock.dumpBlock(self)
    print "  IC Block :: " + self.IC[0].getBlockName() + "@" + self.IC[0].getBlockType()
    
    
class TestBlock(BaseBlock):
  """
  TestBlock represents Test blocks which has two special attributes:
  "TRUE_in" and "FALSE_in". 
  """
  
  def __init__(self, block_name, block_type):
    BaseBlock.__init__(self, block_name, block_type)
    # a tuple of link-in TRUE_in block object and the name of the connection between them
    self.TRUE_in = None
    # a tuple of link-in FALSE_in block object and the name of the connection between them
    self.FALSE_in = None
    
  def __str__(self):    
    return BaseBlock.__str__(self) + \
           "  TRUE_in Block :: " + self.TRUE_in[0].getBlockName() + "@" + \
           self.TRUE_in[0].getBlockType() + "\n" + \
           "  FALSE_in Block :: " + self.FALSE_in[0].getBlockName() + "@" + \
           self.FALSE_in[0].getBlockType() + "\n"
  
  def __repr__(self):    
    return BaseBlock.__repr__(self) + \
           "  TRUE_in Block :: " + self.TRUE_in[0].getBlockName() + "@" + \
           self.TRUE_in[0].getBlockType() + "\n" + \
           "  FALSE_in Block :: " + self.FALSE_in[0].getBlockName() + "@" + \
           self.FALSE_in[0].getBlockType() + "\n"
    
  def dumpBlock(self):
    BaseBlock.dumpBlock(self)
    print "  TRUE_in Block :: " + self.TRUE_in[0].getBlockName() + "@" + \
           self.TRUE_in[0].getBlockType() + "\n" + \
           "  FALSE_in Block :: " + self.FALSE_in[0].getBlockName() + "@" + \
           self.FALSE_in[0].getBlockType()

               
class ChildBlock(BaseBlock):
  """
  ChildBlock represents Child blocks which has a special attribute:
  "model_instance" and "FALSE_in". 
  """
  
  def __init__(self, block_name, block_type, model_instance):
    BaseBlock.__init__(self, block_name, block_type)
    # the instance of the model represented by the child block
    self.model_instance = model_instance
  def __str__(self):
    return BaseBlock.__str__(self) + "  Class Name :: " + str(self.model_instance.__class__) + "\n"
  
  def __repr__(self):
    return BaseBlock.__repr__(self) + "  Class Name :: " + str(self.model_instance.__class__) + "\n"
    
  def dumpBlock(self):
    BaseBlock.dumpBlock(self)
    print "  Class Name :: " + str(self.model_instance.__class__)
    
  def resetChildrenBlocksName(self):
    """
    reset all blocks names of the model represented by the child block in a hierarchical way.
    """
    childrenBlocks = self.model_instance.getBlocks() 
    for childBlock in childrenBlocks:      
      childBlock.resetBlockName(self.getBlockName())
      if isinstance(childBlock, ChildBlock):
        childBlock.resetChildrenBlocksName()
    
    
    
    
    
    
