""" This module implements a dependency graph
    @author: Marc Provost
    @organization: McGill University
    @license: GNU General Public License
    @contact: marc.provost@mail.mcgill.ca
"""

import copy

class DepNode:
  """ Class implementing a node in the dependency graph.
  """
  
  def __init__(self, object):
    """ DepNode's constructor.
        @param object: Reference to a semantic object identifying the node
        @type object: Object
    """
    self.__object = object
    self.__isMarked   = False

  def mark(self):
    self.__isMarked = True
  
  def unMark(self):
    self.__isMarked = False
  
  def isMarked(self):
    return self.__isMarked 

  def getMappedObj(self):
    return self.__object
        
  def __str__(self):
    return "DepNode :: "+str(self.__object)
  
  def __repr__(self):
    return "DepNode :: "+str(self.__object)                 
          
class DepGraph:
  """ Class implementing dependency graph. 
  """

  def __init__(self):
    """ DepGraph's constructor.
    """
    #Dict holding a mapping "Object -> DepNode"
    self.__semanticMapping = {}
    
    #map object->list of objects depending on object
    self.__dependents = {}
    #map object->list of objects that influences object
    self.__influencers = {} 
  
  def addMember(self, object):
    """ Add an object mapped to this graph.
        @param object: the object to be added
        @type object: Object
        @raise ValueError: If object is already in the graph
    """
    if not self.hasMember(object): 
      node = DepNode(object) 
      self.__dependents[object]                     = []
      self.__influencers[object]                    = []
      self.__semanticMapping[object] = node    
    else:
      raise ValueError("Specified object is already member of this graph")
      
  def hasMember(self, object):
    return self.__semanticMapping.has_key(object)      

  def removeMember(self, object):
    """ Remove a object from this graph.
        @param object: the object to be removed
        @type object: Object
        @raise ValueError: If object is not in the graph
    """  
    if self.hasMember(object):
      for dependent in self.getDependents(object):
        self.__influencers[dependent].remove(object)
      for influencer in self.getInfluencers(object):
        self.__dependents[influencer].remove(object)
      
      del self.__dependents[object]
      del self.__influencers[object]
      del self.__semanticMapping[object]       
    else:
      raise ValueError("Specified object is not member of this graph")

  def setDependency(self, dependent, influencer):
    """ Creates a dependency between two objects.
        @param dependent: The object which depends on the other
        @param influcencer: The object which influences the other
        @type dependent: Object
        @type dependent: Object
        @raise ValueError: if depedent or influencer is not member of this graph
        @raise ValueError: if the dependency already exists
    """
    if self.hasMember(dependent) and self.hasMember(influencer): 
      if not influencer in self.__influencers[dependent] and\
         not dependent in self.__dependents[influencer]:
        self.__influencers[dependent].append(influencer) 
        self.__dependents[influencer].append(dependent)
      else:
        raise ValueError("Specified dependency already exists")
    else:
      if not self.hasMember(dependent):
        raise ValueError("Specified dependent object is not member of this graph")
      if not self.hasMember(influencer):
        raise ValueError("Specified influencer object is not member of this graph")
        
  def hasDependency(self, dependent, influencer):
    if self.hasMember(dependent) and self.hasMember(influencer):          
      return influencer in self.__influencers[dependent] and\
             dependent in self.__dependents[influencer]    
    else:
      if not self.hasMember(dependent):
        raise ValueError("Specified dependent object is not member of this graph")
      if not self.hasMember(influencer):
        raise ValueError("Specified influencer object is not member of this graph")        
        
  def unsetDependency(self, dependent, influencer):
    """ Removes a dependency between two objects.
        @param dependent: The object which depends on the other
        @param influcencer: The object which influences the other
        @type dependent: Object
        @type dependent: Object
        @raise ValueError: if depedent or influencer is not member of this graph
        @raise ValueError: if the dependency does not exists
    """      
    if self.hasMember(dependent) and self.hasMember(influencer):   
      if influencer in self.__influencers[dependent] and\
         dependent in self.__dependents[influencer]:    
        self.__influencers[dependent].remove(influencer) 
        self.__dependents[influencer].remove(dependent)
      else:
        raise ValueError("Specified dependency does not exists")
    else:
      if not self.hasMember(dependent):
        raise ValueError("Specified dependent object is not member of this graph")
      if not self.hasMember(influencer):
        raise ValueError("Specified influencer object is not member of this graph")
                                                                                                                              
  def getDependents(self, object):
    if self.hasMember(object):
      return copy.copy(self.__dependents[object])
    else:
      raise ValueError("Specified object is not member of this graph") 

  def getInfluencers(self, object):
    if self.hasMember(object):
      return copy.copy(self.__influencers[object])
    else:
      raise ValueError("Specified object is not member of this graph") 
      
  def getStrongComponents(self):
    return self.__strongComponents()
      
  def __getDepNode(self, object):
    if self.hasMember(object):
      return self.__semanticMapping[object]
    else:
      raise ValueError("Specified object is not a member of this graph")      
      
  def __mark(self, object):
    self.__getDepNode(object).mark()
    
  def __unMark(self, object):
    self.__getDepNode(object).unMark()
    
  def __isMarked(self, object):
    return self.__getDepNode(object).isMarked()    
                                                      
  def __topoSort(self):
    """ Performs a topological sort on the graph.
    """
    for object in self.__semanticMapping.keys():
      self.__unMark(object)
    
    sortedList = []
      
    for object in self.__semanticMapping.keys():
      if not self.__isMarked(object):
        self.__dfsSort(object, sortedList)

    return sortedList
    
  def __dfsSort(self, object, sortedList):
    """ Performs a depth first search collecting
        the objects in topological order.
        @param object: the currently visited object.
        @param sortedList: partial sorted list of objects
        @type object: Object
        @type sortedList: list Of Object
    """
    
    if not self.__isMarked(object):
      self.__mark(object)
    
      for influencer in self.getInfluencers(object):      
        self.__dfsSort(influencer, sortedList)
    
      sortedList.append(object)

  def __strongComponents(self):
    """ Determine the strong components of the graph
        @rtype: list of list of Object
    """
    strongComponents = []
    sortedList = self.__topoSort()
    
    for object in self.__semanticMapping.keys():
      self.__unMark(object) 
          
    sortedList.reverse()
        
    for object in sortedList:
      if not self.__isMarked(object):
        component = []      
        self.__dfsCollect(object, component)
        strongComponents.append(component)
    
    strongComponents.reverse()         
    return strongComponents

  def __dfsCollect(self, object, component):
    """ Collects objects member of a strong component.
        @param object: Node currently visited
        @param component: current component
        @type object: Object
        @type component: List of Object
    """
    if not self.__isMarked(object):
      self.__mark(object)
                  
      for dependent in self.getDependents(object):
        self.__dfsCollect(dependent, component)
        
      component.append(object)
      
if __name__=="__main__":
  print "***Example Usage for a dependency graph***"
  print ""
  g = DepGraph()
  
  print "Adding 3 members to the graph, in this case integers.."
  print "In the simulator, the members will be to model nodes, but"
  print "in general, a member can be any object.."

  g.addMember(1)
  g.addMember(2)
  g.addMember(3)
  
  print "Creating a cycle between the 3 nodes.."
  print ""
  
  g.setDependency(1,2)
  g.setDependency(2,3)
  g.setDependency(3,1)
  
  print "Interface for querying info about dependencies: "
  print ""
  print "getInfluencers(member) : returns the members that the input member depends on"
  print "getDependents(member) : returns the members that depends on the input member"
  print "hasDependency(dependent, influencer) : check if a dependency exists"
  print ""
  print "..see the source file for more useful functions.."
  print ""
  

  print "Getting the strongComponents : "
  print "" 
  print g.getStrongComponents()
  print "" 
  
  print "Breaking the cycle by removing the dependency between 3 and 1: "
  g.unsetDependency(3,1)

  print "To remove a node, use removeMember()"

  print "Getting the strongComponents : "
  print "" 
  print g.getStrongComponents()
  print "" 
    
      
      
      
      
      
      
                            
        